var ExtractionPrep = function()
{
	var exprep = {};
	var debug = 0;
	
	exprep.REAGENTS = [
		'xyleneSpecimen', 'etoh995Specimen', 'bufferPkdSpecimen', 'bufferAtlSpecimen',
		'proteinaseKSpecimen', 'rnaseASpecimen', 'allPrepFfpeKit', 'etoh995',
		'bufferRlt', 'bufferFrn', 'bufferRpe', 'bufferAl',
		'bufferAw1', 'bufferAw2', 'bufferAte', 'bufferRdd',
		'rnaseFreeWater', 'dnaseMix', 'rneasyMinEluteSpinColumn', 'qiaampMinEluteSpinColumn',
		'kapaSybrFast', 'qualityControlPrimersQcpRgt', 'qualityControlTemplateQct', 'qubitDnaHighSens', 'qubitDnaBroadRange'
		];

	var DEFAULT_RNA_VOLUME = 25;
	var DEFAULT_DNA_VOLUME = 25;
	var MINIMAL_PRE_NORMALIZE_QUANTITY = 1.1; // µg
	var EXTRA_QIACUBE_SAMPLE_NAME_START = 'E';
	var protocolMap = {};

	var sourceItemListIsValid = false;
	var qiacubeSortedPrefixes = 'BCADEFGHIJKLMNOPQRSTUVWXYZ';
	var qiacubeDateIsAfterIsolationDate = true;
	var qiacubeRunNoIsValid = false;
	var qiacubePositionsAreValid = false;
	var counterBalanceInfoRow = false;
	var nextStartListName = '';
	var trackingProtocolPreview = true;
	var createdStartList = [];
	
	// Page initialization
	exprep.initPage = function()
	{
		// Step 1
		Events.addEventHandler('step-1', 'wizard-validate', exprep.validateStep1);
		Events.addEventHandler('extractSourceItems', 'change', exprep.sourceItemsOnChange);

		// Step 2
		Events.addEventHandler('step-2', 'wizard-initialize', exprep.initializeStep2);
		Events.addEventHandler('step-2', 'wizard-validate', exprep.validateStep2);
		
		Events.addEventHandler('isolationDate', 'change', Wizard.validateDate);
		Events.addEventHandler('qiacubeDate', 'change', Wizard.validateDate);
		Events.addEventHandler('isolationDate', 'change', exprep.qiacubeDateOnChange);
		Events.addEventHandler('qiacubeDate', 'change', exprep.qiacubeDateOnChange);
		Events.addEventHandler('qiacubeRunNo', 'change', exprep.qiacubeRunNoOnChange);
		
		// Step 3
		Events.addEventHandler('step-3', 'wizard-initialize', exprep.initializeStep3);
		Events.addEventHandler('step-3', 'wizard-validate', exprep.validateStep3);
	
		Buttons.addClickHandler('downloadSampleIdFile', exprep.downloadSampleIdFile);
		Buttons.addClickHandler('downloadQPcrSampleIdFile', exprep.downloadQPcrSampleIdFile);
		
		// Navigation
		Buttons.addClickHandler('gocancel', Wizard.cancelWizard);
		Buttons.addClickHandler('gorestart', Wizard.restartWizard);
		Buttons.addClickHandler('gonext', Wizard.goNextOnClick);
		Buttons.addClickHandler('goregister', Wizard.goRegister);
		Buttons.addClickHandler('gocreate', exprep.createProtocol);

		// Final registration
		Events.addEventHandler('wizard', 'wizard-submit', exprep.submit);

		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetUnprocessedItems';		
		Wizard.showLoadingAnimation('Loading unprocessed items...');
		Wizard.asyncJsonRequest(url, exprep.initializeStep1);
	}
	
	exprep.initializeStep1 = function(response)
	{
		var frm = document.forms['meludi'];
		var specimens = response.specimens;
		var rnas = response.rnas;
		var dnas = response.dnas;
		var items = response.items;
		
		var sourceItemList = frm.extractSourceItems;
		if (items.length)
		{
			var numItems = items.length;
			for (var i=0; i < items.length; i++)
			{
				var item = items[i];
				var name = (i+1) + ': ' + Strings.encodeTags(item.name);
				if (item.bioWell)
				{
					name += ' -- ' + Strings.encodeTags(item.bioWell.bioPlate.name + ' (' + item.bioWell.location + ')');
				}
				var selected = i < 24;
				// Identification of item via id works here as specimens and extracts
				// are stored in the same database table using discriminators, and are
				// therefore guaranteed to not have the same id value as another item in the table.
				var option = new Option(name, item.id, selected, selected);
				option.item = item;
				sourceItemList.options[sourceItemList.length] = option;
			}
			exprep.sourceItemsOnChange();
		}
		else
		{
			Wizard.setFatalError('No items available for processing.');
			return;
		}

		Doc.show('step-1');
		Doc.show('gonext');
	}

	exprep.sourceItemsOnChange = function()
	{
		sourceItemListIsValid = false;
		Wizard.setInputStatus('extractSourceItems');
		
		var frm = document.forms['meludi'];
		var sourceItemList = frm.extractSourceItems;
		
		var numSelected = 0;
		for (var i = 0; i < sourceItemList.length; i++)
		{
			if (sourceItemList[i].selected) numSelected++;
		}

		if (numSelected < 1 || numSelected > 24)
		{
			Wizard.setInputStatus('extractSourceItems', 'invalid', 'Must select 1-24 items.');
			return;
		}
		
		sourceItemListIsValid = true;
		Wizard.setInputStatus('extractSourceItems', 'valid');
	}

	
	exprep.itemNameOnChange = function(event)
	{
		var frm = document.forms['meludi'];
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var numSpecimenItems = exprep.getNumberOfSpecimenItems();
		var numQiacubeItems = numSpecimenItems;
		var totNumItems = numItems;
		exprep.updateDetailsTable();
	}

	exprep.validateStep1 = function(event)
	{
		if (!sourceItemListIsValid)
		{
			event.preventDefault();
		}
	}
	
	exprep.initializeStep2 = function()
	{
		var frm = document.forms['meludi'];
		frm.extractSourceItems.disabled = true;
/*
		frm.rnaDefaultVolume.value = DEFAULT_RNA_VOLUME;
		frm.dnaDefaultVolume.value = DEFAULT_DNA_VOLUME;
*/
		
		Wizard.setCurrentStep(2);
		Doc.show('gocancel');
		Doc.show('gonext');

		Meludi.loadProtocols('SAMPLE_HANDLING_PROTOCOL', 'extractionProtocol');
		Doc.addClass('extractionProtocol', 'list-loading');
		frm.extractionProtocol[0] = new Option('loading...', '');
		var url = '../Protocol.servlet?ID='+App.getSessionId();
		url += '&cmd=GetProtocols&subtype=EXTRACTION_PROTOCOL';
		Wizard.asyncJsonRequest(url, exprep.protocolsLoaded);

		frm.isolationDate.focus();
		Events.sendChangeEvent('isolationDate');
		Events.sendChangeEvent('qiacubeDate');
	}
	
	exprep.protocolsLoaded = function(response)
	{
		var frm = document.forms['meludi'];
		var protocols = response.protocols;

		Doc.removeClass('extractionProtocol', 'list-loading');
		frm.extractionProtocol.length = 0;
		for (var i = 0; i < protocols.length; i++)
		{
			var name = protocols[i].name;
			var id = protocols[i].id;
			protocolMap[id] = name;
			frm.extractionProtocol[frm.extractionProtocol.length] = new Option(name, id, false, true);
		}		
		frm.extractionProtocol[frm.extractionProtocol.length] = new Option('- none -', '');

		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetLastProcessedDna';
		Wizard.asyncJsonRequest(url, exprep.lastProcessedDnaLoaded);
	}
	
	exprep.lastProcessedDnaLoaded = function(response)
	{
		var frm = document.forms['meludi'];
		var dnas = response.dnas;

		if (dnas.length > 0)
		{
			if (dnas.length > 0)
			{
				var dna = dnas[0];

				// Reset the existing values

				// AllPrep FFPE Kit			
				var allPrepFfpeKit = '';

/*
				// Specimen reagents
				var specXylene = '';
				var specEtoh995 = '';
				var specBufferPkd = '';
				var specBufferAtl = '';
				var specProteinaseK = '';
				var specRNAseA = '';

				// QIAcube reagents
				var bufferRlt = '';
				var etoh995 = '';
				var bufferFrn = '';
				var bufferRpe = '';
				var rnaseFreeWater = '';
				var dnaseMix = '';
				var bufferRdd = '';
				var rneasyMinEluteSpinColumn = '';
				var bufferAl = '';
				var bufferAW1 = '';
				var bufferAW2 = '';
				var bufferAte = '';
				var qiaampMinEluteSpinColumn = '';
*/

				if (dna) 
				{
					// AllPrep FFPE Kit			
					allPrepFfpeKit = dna.allPrepFfpeKit;

/*
					// Specimen reagents
					specXylene = dna.xyleneSpecimen;
					specEtoh995 = dna.etoh995Specimen;
					specBufferPkd = dna.bufferPkdSpecimen;
					specBufferAtl = dna.bufferAtlSpecimen;
					specProteinaseK = dna.proteinaseKSpecimen;
					specRNAseA = dna.rnaseASpecimen;

					// QIAcube reagents
					bufferRlt = dna.bufferRlt;
					etoh995 = dna.etoh995;
					bufferFrn = dna.bufferFrn;
					bufferRpe = dna.bufferRpe;
					rnaseFreeWater = dna.rnaseFreeWater;
					dnaseMix = dna.dnaseMix;
					bufferRdd = dna.bufferRdd;
					rneasyMinEluteSpinColumn = dna.rneasyMinEluteSpinColumn;
					bufferAl = dna.bufferAl;
					bufferAW1 = dna.bufferAw1;
					bufferAW2 = dna.bufferAw2;
					bufferAte = dna.bufferAte;
					qiaampMinEluteSpinColumn = dna.qiaampMinEluteSpinColumn;
*/
				}

				// Update reagent values

				// AllPrep FFPE Kit			
				frm.allPrepFfpeKit.value = allPrepFfpeKit;

/*
				// Specimen reagents
				frm.xyleneSpecimen.value = specXylene;
				frm.etoh995Specimen.value = specEtoh995;
				frm.bufferPkdSpecimen.value = specBufferPkd;
				frm.bufferAtlSpecimen.value = specBufferAtl;
				frm.proteinaseKSpecimen.value = specProteinaseK;
				frm.rnaseASpecimen.value = specRNAseA;

				// QIAcube reagents
				frm.etoh995.value = etoh995;
				frm.bufferRlt.value = bufferRlt;
				frm.bufferFrn.value = bufferFrn;
				frm.bufferRpe.value = bufferRpe;
				frm.bufferAl.value = bufferAl;
				frm.bufferAw1.value = bufferAW1;
				frm.bufferAw2.value = bufferAW2;
				frm.bufferAte.value = bufferAte;
				frm.bufferRdd.value = bufferRdd;
				frm.rnaseFreeWater.value = rnaseFreeWater;
				frm.dnaseMix.value = dnaseMix;
				frm.rneasyMinEluteSpinColumn.value = rneasyMinEluteSpinColumn;
				frm.qiaampMinEluteSpinColumn.value = qiaampMinEluteSpinColumn;
*/
			}
		}

		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetNextStartListName';
		Wizard.asyncJsonRequest(url, exprep.startListNameLoaded);
	}
	
	exprep.startListNameLoaded = function(response)
	{
		var frm = document.forms['meludi'];
		var nextStartListName = response.nextStartListName;
		//alert("exprep.startListNameLoaded(): nextStartListName = " + nextStartListName);
		frm.startListName.value = nextStartListName;
	}
	
	exprep.qiacubeDateOnChange = function()
	{
		var frm = document.forms['meludi'];
		
		var isolationDateIsValid = Wizard.isValid('isolationDate');
		var qiacubeDateIsValid = Wizard.isValid('qiacubeDate');
		
		if (!isolationDateIsValid || !qiacubeDateIsValid) return;

		var isolationDate = Dates.parseString(frm.isolationDate.value, 'yyyyMMdd');
		var qiacubeDate = Dates.parseString(frm.qiacubeDate.value, 'yyyyMMdd');
		if (isolationDate > qiacubeDate)
		{
			Wizard.setInputStatus('qiacubeDate', 'invalid', 'QIAcube date is before isolation date');
			qiacubeDateIsAfterIsolationDate = false;
			return;
		}

		qiacubeDateIsAfterIsolationDate = true;

		Wizard.setInputStatus('qiacubeDate', 'valid');
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetNextQiacubeRunNo&qiacubeDate='+encodeURIComponent(frm.qiacubeDate.value);
		Wizard.showLoadingAnimation('Loading Qiacube run number...');
		Wizard.asyncJsonRequest(url, exprep.gotNextRunNumber);
	}

	exprep.gotNextRunNumber = function(response)
	{
		var frm = document.forms['meludi'];
		frm.qiacubeRunNo.value = response.nextQiacubeRunNo;
		Events.sendChangeEvent('qiacubeRunNo');
	}

	exprep.qiacubeRunNoOnChange = function()
	{
		var frm = document.forms['meludi'];
		var runNumber = frm.qiacubeRunNo.value;
		qiacubeRunNoIsValid = false;

		if (runNumber == '')
		{
			Wizard.setInputStatus('qiacubeRunNo', 'invalid', 'Missing');
			return;
		}
		if (!(parseInt(runNumber, 10) >= 1))
		{
			Wizard.setInputStatus('qiacubeRunNo', 'invalid', 'Must be at least 1');
			return;		
		}
		
		Wizard.setInputStatus('qiacubeRunNo', 'valid');
		qiacubeRunNoIsValid = true;
	}

	
	exprep.validateStep2 = function(event)
	{
		var valid = true;
		valid &= Wizard.isValid('isolationDate');
		valid &= Wizard.isValid('qiacubeDate');
		valid &= qiacubeRunNoIsValid;
		
		if (!valid) event.preventDefault();
	}

	
	exprep.initializeStep3 = function()
	{
		var frm = document.forms['meludi'];
		var startListName = frm.startListName.value;
		Doc.element('detailsStartListName').innerHTML = '<b>Start list: ' + startListName + '</b>';
/*
		var rnaDefaultVolume = ExtractUtils.getNumber(frm.rnaDefaultVolume.value);
		var dnaDefaultVolume = ExtractUtils.getNumber(frm.dnaDefaultVolume.value);
*/
		qiacubeSortedPrefixes = exprep.getQiacubeSortedPrefixes();
		
		var html = '<table id="detailsTable" class="step-form">';
		// First header row
		html += '<thead>';
		html += '<tr>';
		html += '<th>Extract</th>';
		html += '<th class="dottedleft">QIAcube</th>';
		html += '<th colspan="1" class="dottedleft"></th>';
		html += '<th colspan="1" class="dottedleft">NanoDrop</th>';
/*
		html += '<th colspan="5" class="dottedleft">RNA</th>';
*/
/*
		html += '<th colspan="3" class="dottedleft">RNA</th>';
		html += '<th colspan="4" class="dottedleft">DNA</th>';
*/
		html += '<th colspan="1" class="dottedleft">RNA</th>';
		html += '<th colspan="1" class="dottedleft">DNA</th>';
		html += '<th class="dottedleft comment">Comment</th>';
		html += '</tr>';
		// Second header row
		html += '<tr>';
		html += '<th>source item</th>';
		html += '<th class="dottedleft">position</th>';
		html += '<td class="dottedleft">Storage</td>';
/*
		html += '<td class="dottedleft">Pos</td><td>Vol.</td><td>Conc.</td><td>Quantity</td><td><span id="toggle-normalize-1" class="link" title="Toggle all">Pre-</span></td>';
*/
/*
		html += '<td class="dottedleft">Pos</td><td>Vol.</td><td>Conc.</td><td>Quantity</td><td>Pre-</td>';
*/
/*
		html += '<td class="dottedleft">Pos</td><td>Vol.</td><td>Conc.</td><td>Quantity</td>';
		html += '<td class="dottedleft">Vol.</td><td>Conc.</td><td>Quantity</td><td>ΔCt</td>';
*/
		html += '<th class="dottedleft">Plate Pos.</th>';
		html += '<th class="dottedleft">Plate Pos.</th>';
		html += '<th class="dottedleft">Plate Pos.</th>';
		html += '<td class="dottedleft comment"></td>';
		html += '</tr>';
		// Third header row
		html += '<tr>';
		html += '<th></th>';
		html += '<th class="dottedleft"></th>';
		html += '<td class="dottedleft">box</td>';
/*
		html += '<td class="dottedleft"></td><td>(µl)</td><td>(ng/µl)</td><td>(µg)</td><td><span id="toggle-normalize-2" class="link" title="Toggle all">normalize</span></td>';
*/
/*
		html += '<td class="dottedleft"></td><td>(µl)</td><td>(ng/µl)</td><td>(µg)</td><td>normalize</td>';
*/
/*
		html += '<td class="dottedleft"></td><td>(µl)</td><td>(ng/µl)</td><td>(µg)</td>';
		html += '<td class="dottedleft">(µl)</td><td>(ng/µl)</td><td>(µg)</td><td></td>';
*/
		html += '<td class="dottedleft"></td>';
		html += '<td class="dottedleft"></td>';
		html += '<td class="dottedleft"></td>';
		html += '<td class="dottedleft comment"></td>';
		html += '</tr>';
		html += '</thead>';
		html += '<tbody>';

		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var numSpecimenItems = exprep.getNumberOfSpecimenItems();
		var numRnaItems = exprep.getNumberOfRnaItems();
		var numDnaItems = exprep.getNumberOfDnaItems();
		var numNanoDropItems = numSpecimenItems + numRnaItems;
		var numQiacubeItems = numSpecimenItems;
		var totNumItems = numItems;
		var itemNo = 0;
		var qiacubeItemNo = 0;
		// Set default values for DNA/RNA extraction
		for (var i = 0; i < totNumItems; i++)
		{
			itemNo++;
			var item = null;
			var storageBox = '-';
			var rnaStorageBox = '-';
			var dnaStorageBox = '-';
			if (i < numItems)
			{
				item = selItemsList[i];
				var qiacubePos = null;
				if (ExtractUtils.isSpecimen(item))
				{
					qiacubeItemNo++;
					qiacubePos = ExtractUtils.getDetailedQiaCubePosition(numQiacubeItems, qiacubeItemNo, qiacubeSortedPrefixes);
				}
				item.qiacubePosition = qiacubePos;
				item.originalQiacubePosition = qiacubePos;
				
				var rna = {};
				var dna = {};
				// RNA
				rna.name = item.name + '.r';
/*
				rna.volume = rnaDefaultVolume;
*/
				// DNA
				dna.name = item.name + '.d';
/*
				dna.volume = dnaDefaultVolume;
*/
/*				
				item.rna = rna;
				item.dna = dna;
*/
				if (!item.rna)
				{
					item.rna = rna;
				}
				if (!item.dna)
				{
					item.dna = dna;
				}
/*
				var storageBox = '-';
*/
				// Input extract source item specimen/RNA/DNA
				if (item.bioWell != null)
				{
					storageBox = Strings.encodeTags(item.bioWell.bioPlate.name + '[' + item.bioWell.location + ']');
				}
				// RNA extract from specimen
				if (item.rna && item.rna.bioWell)
				{
					rnaStorageBox = Strings.encodeTags(item.rna.bioWell.bioPlate.name + '[' + item.rna.bioWell.location + ']');
				}
				// DNA extract from specimen
				if (item.dna && item.dna.bioWell)
				{
					dnaStorageBox = Strings.encodeTags(item.dna.bioWell.bioPlate.name + '[' + item.dna.bioWell.location + ']');
				}
				// Input RNA item storage location should be displayed in the RNA extract storage location column
				if (ExtractUtils.isRna(item) && item.bioWell != null)
				{
					rnaStorageBox = Strings.encodeTags(item.bioWell.bioPlate.name + '[' + item.bioWell.location + ']');
				}
				// Input DNA item storage location should be displayed in the DNA extract storage location column
				if (ExtractUtils.isDna(item) && item.bioWell != null)
				{
					dnaStorageBox = Strings.encodeTags(item.bioWell.bioPlate.name + '[' + item.bioWell.location + ']');
				}
			}
				
			html += '<tr class="highlight">';
			if (i < numItems)
			{
				html += '<th>' + Strings.encodeTags(item.name) + '</th>';
			}
			else
			{
				// Extra QIAcube item
				html += '<th><input type="text" id="itemName'+i+'" style="font-weight:bold;" value="' + Strings.encodeTags(item.name) + '"></intput></th>';
			}
			html += '<td id="qiacubePosition'+i+'" class="dottedleft"></td>';
			html += '<td class="dottedleft">'+storageBox+'</td>';
			html += '<td id="rnaPos'+i+'" class="dottedleft"></td>';
/*
			html += '<td id="rnaVol'+i+'"></td>';
			html += '<td id="rnaConc'+i+'" class="bg-filled-50 italic"></td>';
			html += '<td id="rnaQuantity'+i+'" class="bg-filled-50 italic"></td>';
*/
/*
			html += '<td id="rnaPreNormalize'+i+'"name="rnaPreNormalize'+i+'"></td>';
*/
/*
			html += '<td id="dnaVol'+i+'"></td>';
			html += '<td id="dnaConc'+i+'">-</td>';
			html += '<td id="dnaQuantity'+i+'" class="bg-filled-50 italic"></td>';
			html += '<td id="deltaCt'+i+'" class="bg-filled-50 italic">-</td>';
*/
			html += '<td id="rnaPlatePos'+i+'" class="dottedleft">'+rnaStorageBox+'</td>';
			html += '<td id="dnaPlatePos'+i+'" class="dottedleft">'+dnaStorageBox+'</td>';
			html += '<td id="comments'+i+'" class="dottedleft comment"></td>';
			html += '</tr>';
		}
		if (numQiacubeItems == 1 || numQiacubeItems == 11 || numQiacubeItems == 23)
		{
			// Add extra row for counterbalance in QIAcube
			counterBalanceInfoRow = true;
			var i = numItems;
			var qiacubePos = ExtractUtils.getDetailedQiaCubePosition(numQiacubeItems + 1, numQiacubeItems + 1, qiacubeSortedPrefixes);
				
			html += '<tr class="highlight">';
			html += '<th>' + Strings.encodeTags('COUNTER-BALANCE') + '</th>';
			html += '<td id="qiacubePosition'+i+'" class="dottedleft">'+qiacubePos+'</td>';
			html += '<td class="dottedleft">-</td>';
			html += '<td id="rnaPos'+i+'" class="dottedleft">-</td>';
/*
			html += '<td id="rnaVol'+i+'">-</td>';
			html += '<td id="rnaConc'+i+'" class="bg-filled-50 italic">-</td>';
			html += '<td id="rnaQuantity'+i+'" class="bg-filled-50 italic">-</td>';
*/
/*
			html += '<td>-</td>';
*/
/*
			html += '<td id="dnaVol'+i+'">-</td>';
			html += '<td id="dnaConc'+i+'">-</td>';
			html += '<td id="dnaQuantity'+i+'" class="bg-filled-50 italic">-</td>';
			html += '<td id="deltaCt'+i+'" class="bg-filled-50 italic">-</td>';
*/
			html += '<td id="rnaPlatePos'+i+'" class="dottedleft">-</td>';
			html += '<td id="dnaPlatePos'+i+'" class="dottedleft">-</td>';
			html += '<td id="comments'+i+'" class="dottedleft comment">Reminder to use counterbalance</td>';
			html += '</tr>';				
		}
		html += '</tbody>';
		html += '</table>';
		Doc.element('detailsSection').innerHTML = html;

		for (var i = numItems; i < totNumItems; i++)
		{
			Events.addEventHandler('itemName'+i, 'change', exprep.itemNameOnChange);
		}

		exprep.updateDetailsTable();
		
		Wizard.setCurrentStep(3);
		Doc.show('gocancel');
		Doc.show('goregister');
		Doc.show('gocreate');
		//Doc.hide('gocreate');

		Wizard.keepSessionAlive();
		
		if (debug && location.protocol != 'https:') 
		{
			Wizard.showDebugTools();
		}
	}

	exprep.getQiacubeSortedPrefixes = function()
	{
		var frm = document.forms['meludi'];
		var primaryQiacubePrefix = frm.qiacubePrimaryPrefix.value;
		var secondaryQiacubePrefix = frm.qiacubeSecondaryPrefix.value;

		// Get string of current QIAcube prefixes in order
		var sortedPrefixes = ExtractUtils.getQiacubeSortedPrefixes(primaryQiacubePrefix, secondaryQiacubePrefix);
		return sortedPrefixes;
	}

	exprep.getSelectedItemsList = function()
	{
		var frm = document.forms['meludi'];
		
		// Get number of selected extract source items
		var numItems = 0;
		var selItemsList = [];
		for (var i = 0; i < frm.extractSourceItems.length; i++)
		{
			if (frm.extractSourceItems[i].selected)
			{
				if (frm.extractSourceItems[i].item != null)
				{
					selItemsList[numItems] = frm.extractSourceItems[i].item;
					numItems++;
				}
			}
		}
		return selItemsList;
	}

	exprep.getNumberOfSpecimenItems = function()
	{
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var numSpecimen = 0;
		for (var i = 0; i < numItems; i++)
		{
			var item = selItemsList[i];
			if (ExtractUtils.isSpecimen(item))
			{
				numSpecimen++;
			}
		}
		return numSpecimen;
	}
	
	exprep.getNumberOfRnaItems = function()
	{
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var numRna = 0;
		for (var i = 0; i < numItems; i++)
		{
			var item = selItemsList[i];
			if (ExtractUtils.isRna(item))
			{
				numRna++;
			}
		}
		return numRna;
	}
	
	exprep.getNumberOfDnaItems = function()
	{
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var numDna = 0;
		for (var i = 0; i < numItems; i++)
		{
			var item = selItemsList[i];
			if (ExtractUtils.isDna(item))
			{
				numDna++;
			}
		}
		return numDna;
	}

	exprep.updateDetailsTable = function()
	{
		var frm = document.forms['meludi'];
/*
		var rnaDefaultVolume = ExtractUtils.getNumber(frm.rnaDefaultVolume.value);
		var dnaDefaultVolume = ExtractUtils.getNumber(frm.dnaDefaultVolume.value);
*/
		var hasQiacubePosition = [];
		var duplicatePosition = 0;

		qiacubeSortedPrefixes = exprep.getQiacubeSortedPrefixes();
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var totNumItems = numItems;
		// Get sorted list of used integer QIAcube positions for specimens, generalized for more than one QIAcube
		var specSortedIntQiacubePosition = [];
		var qiacubeItemNo = 0;
		for (var i = 0; i < totNumItems; i++)
		{
			if (i < numItems)
			{
				var item = selItemsList[i];
				if (ExtractUtils.isSpecimen(item))
				{
					var intQiacubePosition = ExtractUtils.getIntQiacubePosition(item.qiacubePosition, qiacubeSortedPrefixes);
					specSortedIntQiacubePosition[qiacubeItemNo] = intQiacubePosition;
					qiacubeItemNo++;
				}
			}
		}
		// Sort for ascending integer values
		specSortedIntQiacubePosition.sort(function(a,b){return a-b});
		var numSpecimenItems = exprep.getNumberOfSpecimenItems();
		var numRnaItems = exprep.getNumberOfRnaItems();
		var numDnaItems = exprep.getNumberOfDnaItems();
		var numNanoDropItems = numSpecimenItems + numRnaItems;
		var rnaItemNo = 0;
		// Set default values for DNA/RNA extraction
		for (var i = 0; i < totNumItems; i++)
		{
			var item = null;
			if (i < numItems)
			{
				item = selItemsList[i];
			}
			Doc.element('qiacubePosition'+i).innerHTML = item.qiacubePosition;
			Doc.addOrRemoveClass('qiacubePosition'+i, 'nondefault', item.qiacubePosition != item.originalQiacubePosition);

			var intQiacubePosition = ExtractUtils.getIntQiacubePosition(item.qiacubePosition, qiacubeSortedPrefixes);
			if (intQiacubePosition != null)
			{
				if (hasQiacubePosition[intQiacubePosition])
				{
					duplicatePosition = intQiacubePosition;
				}
				else
				{
					hasQiacubePosition[intQiacubePosition] = true;
				}
			}
			var intNanoDropPlatePosition = intQiacubePosition;
			// Get integer NanoDrop position for specimens from index in sorted list of used integer QIAcube positions
			if (intQiacubePosition)
			{
				var sortIndex = -1;
				for (var j = 0; j < specSortedIntQiacubePosition.length; j++)
				{
					if (intQiacubePosition == specSortedIntQiacubePosition[j])
					{
						sortIndex = j;
					}
				}
				if (sortIndex > -1)
				{
					intNanoDropPlatePosition = sortIndex + 1;
				}
			}
			if (ExtractUtils.isRna(item))
			{
				rnaItemNo++;
				intNanoDropPlatePosition = numSpecimenItems + rnaItemNo;
			}
			
			// Calculate NanoDrop well from Qiacube position
			var row;
			var col;
			// Specimens from up to two QIAcube instruments and input RNA items placed on one NanoDrop plate
			if (intNanoDropPlatePosition > 16)
			{
				// (17 - 24) NanoDrop plate for QIAcube 'A', positions A3 - H3
				row = intNanoDropPlatePosition - 17;
				col = 3;
			}
			else if (intNanoDropPlatePosition > 8)
			{
				// ( 9 - 12) NanoDrop plate for QIAcube 'C', positions A2 - D2
				// (13 - 16) NanoDrop plate for QIAcube 'A', positions E2 - H2
				row = intNanoDropPlatePosition - 9;
				col = 2;
			}
			else if (intNanoDropPlatePosition > 0)
			{
				// ( 1 -  8) NanoDrop plate for QIAcube 'C', positions A1 - H1
				row = intNanoDropPlatePosition - 1;
				col = 1;
			}
			if (i < numItems)
			{
				if (!ExtractUtils.isDna(item))
				{
					//item.rna.well = Meludi.wellToAlpha(row) + col;
					item.rna.nanoDropWell = Meludi.wellToAlpha(row) + col;
				}
				else
				{
					//item.dna.well = null;
					item.rna.nanoDropWell = null;
				}

/*
				Doc.element('dnaVol'+i).innerHTML = item.dna.volume;
				Doc.addOrRemoveClass('dnaVol'+i, 'nondefault', item.dna.volume != dnaDefaultVolume);
*/
/*
				var dnaConc = Doc.element('dnaConc'+i);
				dnaConc.value = Meludi.formatNumber(item.dna.ndConc) || '';
				dnaConc.title = '260/280=' + item.dna.nd260by280 + '\n260/230=' + item.dna.nd260by230;
				if (dnaConc != null)
				{
					var dnaConcValue = dnaConc.value;
					if (dnaConcValue != null && dnaConcValue != '-' && dnaConcValue != '')
					{
						item.dna.ndConc = ExtractUtils.getNumber(dnaConcValue);
					}
				}
				item.dna.quantity = item.dna.qubitConc * item.dna.volume / 1000;
				Doc.element('dnaQuantity'+i).innerHTML = Meludi.formatNumber(item.dna.quantity || null, null, 2) || '-';

				var deltaCt = Doc.element('deltaCt'+i);
				if (deltaCt != null)
				{
					deltaCt.innerHTML = Meludi.formatNumber(item.dna.deltaCt) || '-';
				}
*/			
				//Doc.element('rnaPos'+i).innerHTML = item.rna.well;
				Doc.element('rnaPos'+i).innerHTML = item.rna.nanoDropWell;
				Doc.addOrRemoveClass('rnaPos'+i, 'nondefault', item.qiacubePosition != item.originalQiacubePosition);
/*
				Doc.element('rnaVol'+i).innerHTML = item.rna.volume;
				Doc.addOrRemoveClass('rnaVol'+i, 'nondefault', item.rna.volume != rnaDefaultVolume);
				var rnaConc = Doc.element('rnaConc'+i);
				rnaConc.innerHTML = Meludi.formatNumber(item.rna.ndConc) || '-';
				rnaConc.title = '260/280=' + item.rna.nd260by280 + '\n260/230=' + item.rna.nd260by230;
				
				item.rna.quantity = item.rna.ndConc * item.rna.volume / 1000;
				var enoughQuantity = item.rna.quantity >= MINIMAL_PRE_NORMALIZE_QUANTITY;
				item.rna.flag = enoughQuantity ? null : 'NotEnoughRemainingQuantity';

				Doc.element('rnaQuantity'+i).innerHTML = Meludi.formatNumber(item.rna.quantity || null, null, 2) || '-';
*/
				var cmnt = Doc.element('comments'+i);
				var tooltip = [];
				var comment = [];
				if (item.operatorDeliveryComment || item.comment || item.dna.comment || item.rna.comment)
				{
					// Display operator delivery comments directly in the comment field
					if (item.operatorDeliveryComment)
					{
						comment[0] = item.operatorDeliveryComment;
						tooltip[tooltip.length] = 'Op.Del.Cmt: ' + item.operatorDeliveryComment;
					}
					else
					{
						comment[0] = 'Yes';
					}
					if (item.comment) tooltip[tooltip.length] = 'Specimen: ' + item.comment;
					if (item.rna.comment) tooltip[tooltip.length] = 'RNA: ' + item.rna.comment;
					if (item.dna.comment) tooltip[tooltip.length] = 'DNA: ' + item.dna.comment;
				}
				
				var modifiedReagents = false;
				for (var rgNo = 0; rgNo < exprep.REAGENTS.length; rgNo++)
				{
					var rg = exprep.REAGENTS[rgNo];
					if (item[rg]) 
					{
						tooltip[tooltip.length] = Doc.element(rg+'.title').innerHTML + ': ' + Strings.encodeTags(item[rg]);
						modifiedReagents = true;
					}
				}
				if (modifiedReagents)
				{
					comment[comment.length] = 'Reagents';
				}
				
				if (tooltip.length > 0)
				{
					cmnt.innerHTML = comment.join(', ');
					Doc.addClass(cmnt, 'nondefault');
					cmnt.title = tooltip.join('\n');
				}
				else
				{
					cmnt.innerHTML = '';
					Doc.removeClass(cmnt, 'nondefault');
					cmnt.title = null;
				}
			}
			else
			{
				// Extra QIAcube item
				var itemNameEl = Doc.element('itemName'+i);
				if (itemNameEl != null)
				{
					itemNameEl.value = Strings.encodeTags(item.name);
				}
			}
		}

		if (duplicatePosition)
		{
			var duplicatePositionStr = ExtractUtils.getDetailedQiacubePosition(duplicatePosition, qiacubeSortedPrefixes);
			Wizard.setInputStatus('downloadSampleIdFile', 'invalid', 'Duplicate QIAcube position: ' + duplicatePositionStr);
			qiacubePositionsAreValid = false;
		}
		else
		{
			Wizard.setInputStatus('downloadSampleIdFile');
			qiacubePositionsAreValid = true;
		}
	}

	exprep.downloadSampleIdFile = function()
	{
		var frm = document.forms['meludi'];

		if (!qiacubePositionsAreValid)
		{
			Forms.showNotification('downloadSampleIdFile', 'QIAcube positions are not valid. Please fix before downloading the file.');
			return;
		}

		var itemNames = [];
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=DownloadNanoDropSampleIdFile';
		url += '&startListName='+encodeURIComponent(frm.startListName.value);
		url += '&qiacubeDate='+encodeURIComponent(frm.qiacubeDate.value);
		url += '&qiacubeRunNo='+encodeURIComponent(frm.qiacubeRunNo.value);
		url += '&blankControlColumn='+encodeURIComponent(frm.blankControlColumn.checked);
		
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var itemNo = 0;
		for (var i = 0; i < numItems; i++)
		{
			var item = selItemsList[i];
			if (ExtractUtils.isSpecimen(item) || ExtractUtils.isRna(item))
			{
				itemNames[itemNo] = item.name;
				itemNo++;
			}
		}
		url += '&itemNames=' + encodeURIComponent(itemNames.join(','));
		window.open(url);
	}

	exprep.downloadQPcrSampleIdFile = function()
	{
		var frm = document.forms['meludi'];
		var itemNames = [];
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=DownloadQPcrSampleIdFile';
		url += '&startListName='+encodeURIComponent(frm.startListName.value);
		url += '&qiacubeDate='+encodeURIComponent(frm.qiacubeDate.value);
		url += '&qiacubeRunNo='+encodeURIComponent(frm.qiacubeRunNo.value);

		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var itemNo = 0;
		for (var i = 0; i < numItems; i++)
		{
			var item = selItemsList[i];
			if (ExtractUtils.isSpecimen(item) || ExtractUtils.isDna(item))
			{
				itemNames[itemNo] = item.name;
				itemNo++;
			}
		}
		url += '&itemNames=' + encodeURIComponent(itemNames.join(','));
		window.open(url);
	}

	exprep.validateStep3 = function(event)
	{
		var valid = true;
		valid &= qiacubePositionsAreValid;
		
		if (!valid) event.preventDefault();
	}

	exprep.submit = function()
	{
		//alert("exprep.submit(): Start");
		var frm = document.forms['meludi'];
		
		var submitInfo = {};
		submitInfo.items = [];
		submitInfo.extraItems = [];

		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var totNumItems = numItems;
		for (var i = 0; i < totNumItems; i++)
		{
			var item = null;
			if (i < numItems)
			{
				item = selItemsList[i];
				submitInfo.items[submitInfo.items.length] = item;
			}
		}
		
		// Extraction
		submitInfo.startListName = frm.startListName.value;
		submitInfo.isolationDate = frm.isolationDate.value;
		submitInfo.extractionProtocol = parseInt(frm.extractionProtocol.value, 10);
		var extractionProtocolName = '';
		var extractionProtocol = frm.extractionProtocol.value;
		if (extractionProtocol)
		{
			extractionProtocolName = protocolMap[extractionProtocol];
		}
		submitInfo.extractionProtocolName = extractionProtocolName;
/*
		submitInfo.rnaDefaultVolume = ExtractUtils.getNumber(frm.rnaDefaultVolume.value);
		submitInfo.dnaDefaultVolume = ExtractUtils.getNumber(frm.dnaDefaultVolume.value);
*/

		// QIAcube
		submitInfo.qiacubeDate = frm.qiacubeDate.value;
		submitInfo.qiacubePrimaryPrefix = frm.qiacubePrimaryPrefix.value;
		submitInfo.qiacubeSecondaryPrefix = frm.qiacubeSecondaryPrefix.value;
		submitInfo.qiacubeRunNo = parseInt(frm.qiacubeRunNo.value, 10);
		submitInfo.qiacubeOperator = frm.qiacubeOperator.value;
		submitInfo.allPrepFfpeKit = frm.allPrepFfpeKit.value;

/*
		// Reagents
		for (var i = 0; i < exprep.REAGENTS.length; i++)
		{
			var rg = exprep.REAGENTS[i];
			submitInfo[rg] = frm[rg].value;
		}
*/
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=RegisterStartItemList';
		//alert("exprep.submit(): url = " + url);
		Wizard.showLoadingAnimation('Performing registration...');
		Wizard.asyncJsonRequest(url, exprep.submissionResults, 'POST', JSON.stringify(submitInfo));
	}
	
	exprep.submissionResults = function(response)
	{
		Wizard.showFinalMessage(response.messages);
		Doc.show('gorestart');
		Doc.show('gocreate');
		trackingProtocolPreview = false;
		if (response.createdStartList && response.createdStartList.length > 0)
		{
			createdStartList = response.createdStartList;
		}
	}

	exprep.createProtocol = function()
	{
		var frm = document.forms['meludi'];
		var submitInfo = {};
		submitInfo.items = [];

		submitInfo.docMode = 'protocol';
		// Get list of selected extract source items
		var selItemsList = exprep.getSelectedItemsList();
		var numItems = selItemsList.length;
		var totNumItems = numItems;
		for (var i = 0; i < totNumItems; i++)
		{
			var item = null;
			if (i < numItems)
			{
				item = selItemsList[i];
			}
			submitInfo.items[submitInfo.items.length] = item;
		}

		// Extraction
		submitInfo.startListName = frm.startListName.value;
		submitInfo.isolationDate = frm.isolationDate.value;
/*
		submitInfo.extractionProtocol = parseInt(frm.extractionProtocol.value, 10);
*/
		var extractionProtocolName = '';
		var extractionProtocol = frm.extractionProtocol.value;
		if (extractionProtocol)
		{
			extractionProtocolName = protocolMap[extractionProtocol];
		}
		var protocolVersionNo = '';
		if (extractionProtocolName)
		{
			protocolVersionNo = extractionProtocolName;
			// Remove part of string up to and including last "v"
			var vIndex = protocolVersionNo.indexOf('v');
			while (vIndex >= 0)
			{
				// Remove part of string up to and including first "v"
				protocolVersionNo = protocolVersionNo.substring(vIndex + 1);
				vIndex = protocolVersionNo.indexOf('v');
			}
		}
		var protocolPreviewFlag = null;
		if (trackingProtocolPreview)
		{
			protocolPreviewFlag = 'true';
		}
		submitInfo.extractionProtocolName = extractionProtocolName;
		submitInfo.protocolVersionNo = protocolVersionNo;
		submitInfo.protocolPreviewFlag = protocolPreviewFlag;
/*
		submitInfo.dnaDefaultVolume = ExtractUtils.getNumber(frm.dnaDefaultVolume.value);
*/
		submitInfo.dnaDefaultVolume = null;
		
		// QIAcube
		submitInfo.qiacubeDate = frm.qiacubeDate.value;
		submitInfo.qiacubeRunNo = parseInt(frm.qiacubeRunNo.value, 10);
		submitInfo.qiacubeOperator = frm.qiacubeOperator.value;
		submitInfo.allPrepFfpeKit = frm.allPrepFfpeKit.value;

		// Save values in hidden input fields for retrieval after HTML request
		var submitInfoJsonStr = JSON.stringify(submitInfo);
		frm.hiddenSubmitInfo.value = submitInfoJsonStr;
		frm.hiddenNumItems.value = submitInfo.items.length;
		frm.hiddenCounterBalanceInfoRow.value = counterBalanceInfoRow;

		// Calling frm.submit() will change jsp file to extraction_protocol2.jsp
		frm.submit();
	}

	return exprep;
}();

Doc.onLoad(ExtractionPrep.initPage);

